"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _DiskWatcher_registry, _DiskWatcher_options, _DiskWatcher_chunks;
Object.defineProperty(exports, "__esModule", { value: true });
exports.DiskWatcher = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const abstract_1 = require("./abstract");
const util_1 = require("util");
const existing_chunk_1 = require("../chunk/existing-chunk");
const cuid_1 = __importDefault(require("cuid"));
const errors_1 = require("../errors");
const fsExistsAsync = (0, util_1.promisify)(fs_1.default.exists);
const fsMkdirAsync = (0, util_1.promisify)(fs_1.default.mkdir);
const fsReadFileAsync = (0, util_1.promisify)(fs_1.default.readFile);
const fsWriteFileAsync = (0, util_1.promisify)(fs_1.default.writeFile);
const fsAppendFileAsync = (0, util_1.promisify)(fs_1.default.appendFile);
const fsUnlinkAsync = (0, util_1.promisify)(fs_1.default.unlink);
const fsReaddirAsync = (0, util_1.promisify)(fs_1.default.readdir);
class DiskWatcher extends abstract_1.DataWatcher {
    constructor(registry, options) {
        super(options);
        _DiskWatcher_registry.set(this, void 0);
        _DiskWatcher_options.set(this, void 0);
        _DiskWatcher_chunks.set(this, void 0);
        __classPrivateFieldSet(this, _DiskWatcher_registry, registry, "f");
        __classPrivateFieldSet(this, _DiskWatcher_options, options, "f");
        /**
         * Runtime cache storage
         */
        __classPrivateFieldSet(this, _DiskWatcher_chunks, {}, "f");
    }
    async save(saveContract) {
        if (!saveContract.insertRows.length) {
            throw new errors_1.CacheError("E_EMPTY_SAVE" /* E_CODES.E_EMPTY_SAVE */);
            return;
        }
        if (!this.$isWriteable()) {
            await this.$toBeUnblocked();
        }
        const operationId = (0, cuid_1.default)();
        saveContract.chunkRef.setConsistency(false);
        if (!__classPrivateFieldGet(this, _DiskWatcher_chunks, "f")[saveContract.chunkRef.id]) {
            __classPrivateFieldGet(this, _DiskWatcher_chunks, "f")[saveContract.chunkRef.id] = {
                ref: saveContract.chunkRef,
                operations: {}
            };
        }
        __classPrivateFieldGet(this, _DiskWatcher_chunks, "f")[saveContract.chunkRef.id].operations[operationId] = saveContract.insertRows;
        const chunkDirectoryExists = await fsExistsAsync(__classPrivateFieldGet(this, _DiskWatcher_options, "f").disk.outputDirectory);
        if (!chunkDirectoryExists) {
            await fsMkdirAsync(__classPrivateFieldGet(this, _DiskWatcher_options, "f").disk.outputDirectory);
        }
        const chunkFilename = `${saveContract.chunkRef.id}.txt`;
        const chunkPathname = path_1.default.resolve(__classPrivateFieldGet(this, _DiskWatcher_options, "f").disk.outputDirectory, chunkFilename);
        const chunkExists = await fsExistsAsync(chunkPathname);
        if (!chunkExists) {
            const metadata = {
                table: saveContract.chunkRef.table,
                expiresAt: saveContract.chunkRef.expiresAt
            };
            await fsWriteFileAsync(chunkPathname, `${JSON.stringify(metadata)}\n`);
        }
        /**
         * Need some kind of schema to optimize
         */
        const storeData = saveContract.insertRows
            .map(row => JSON.stringify(row))
            .join('\n')
            .concat('\n');
        await fsAppendFileAsync(chunkPathname, storeData)
            .then(() => {
            __classPrivateFieldGet(this, _DiskWatcher_registry, "f").increaseSize(saveContract.chunkRef.id, saveContract.insertRows.length);
            saveContract.chunkRef.setConsistency(true);
            delete __classPrivateFieldGet(this, _DiskWatcher_chunks, "f")[saveContract.chunkRef.id].operations[operationId];
        });
    }
    async load(chunkId) {
        const chunkFilename = `${chunkId}.txt`;
        const chunkPathname = path_1.default.resolve(__classPrivateFieldGet(this, _DiskWatcher_options, "f").disk.outputDirectory, chunkFilename);
        const data = await fsReadFileAsync(chunkPathname, { encoding: 'utf8' });
        const [strMetadata, ...strRows] = data.trim().split('\n');
        const metadata = JSON.parse(strMetadata);
        const rows = strRows.map(strRow => JSON.parse(strRow));
        return {
            chunkRef: new existing_chunk_1.ExistingChunk(this, {
                id: chunkId,
                table: metadata.table,
                expiresAt: metadata.expiresAt,
                size: rows.length
            }),
            loadedRows: rows
        };
    }
    backupRuntimeCache() {
        for (const chunkId in __classPrivateFieldGet(this, _DiskWatcher_chunks, "f")) {
            const chunkDirectoryExists = fs_1.default.existsSync(__classPrivateFieldGet(this, _DiskWatcher_options, "f").disk.outputDirectory);
            if (!chunkDirectoryExists) {
                fs_1.default.mkdirSync(__classPrivateFieldGet(this, _DiskWatcher_options, "f").disk.outputDirectory);
            }
            const chunkFilename = `${chunkId}.txt`;
            const chunkPathname = path_1.default.resolve(__classPrivateFieldGet(this, _DiskWatcher_options, "f").disk.outputDirectory, chunkFilename);
            const chunk = __classPrivateFieldGet(this, _DiskWatcher_chunks, "f")[chunkId].ref;
            const chunkExists = fs_1.default.existsSync(chunkPathname);
            if (!chunkExists) {
                const metadata = {
                    table: chunk.table,
                    expiresAt: chunk.expiresAt
                };
                fs_1.default.writeFileSync(chunkPathname, `${JSON.stringify(metadata)}\n`);
            }
            for (const operationId in __classPrivateFieldGet(this, _DiskWatcher_chunks, "f")[chunkId].operations) {
                const runtimeRows = __classPrivateFieldGet(this, _DiskWatcher_chunks, "f")[chunkId].operations[operationId];
                /**
                 * Need some kind of schema to optimize
                 */
                const storeData = runtimeRows
                    .map(row => JSON.stringify(row))
                    .join('\n')
                    .concat('\n');
                fs_1.default.appendFileSync(chunkPathname, storeData);
                chunk.setConsistency(true);
                delete __classPrivateFieldGet(this, _DiskWatcher_chunks, "f")[chunkId].operations[operationId];
                console.log(`Succesfully backed up operation ${operationId} of chunk ${chunkId} with ${runtimeRows.length} rows`);
            }
        }
    }
    async restore() {
        const chunkDirectoryExists = await fsExistsAsync(__classPrivateFieldGet(this, _DiskWatcher_options, "f").disk.outputDirectory);
        if (!chunkDirectoryExists) {
            return;
        }
        const files = await fsReaddirAsync(__classPrivateFieldGet(this, _DiskWatcher_options, "f").disk.outputDirectory);
        for (const filename of files) {
            const isChunkFile = filename.includes('.txt');
            if (!isChunkFile) {
                continue;
            }
            const chunkId = filename.split('.txt').join('');
            const loaded = await this.load(chunkId);
            __classPrivateFieldGet(this, _DiskWatcher_registry, "f").register(loaded.chunkRef);
        }
    }
    async cleanup(chunkId) {
        const chunkFilename = `${chunkId}.txt`;
        const chunkPathname = path_1.default.resolve(__classPrivateFieldGet(this, _DiskWatcher_options, "f").disk.outputDirectory, chunkFilename);
        await fsUnlinkAsync(chunkPathname);
    }
    countRows(chunkId) {
        return __classPrivateFieldGet(this, _DiskWatcher_registry, "f").getOne(chunkId).size;
    }
}
exports.DiskWatcher = DiskWatcher;
_DiskWatcher_registry = new WeakMap(), _DiskWatcher_options = new WeakMap(), _DiskWatcher_chunks = new WeakMap();
